from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Literal, Any
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule



class WhatsAppVariable(BaseModel):
    placeholder: str = Field(alias="placeholder")
    sample_value: str = Field(alias="sampleValue")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class WhatsAppTemplateRequest(BaseModel):
    name: str = Field(alias="name")
    label: str = Field(alias="label")
    description: str = Field(alias="description")
    category: str = Field(alias="category")
    segment: str = Field(alias="segment")
    periodicity: str = Field(alias="periodicity")
    client_id: str = Field(alias="clientId")
    language: Optional[str] = Field(alias="language")
    header_type: Optional[Literal['none', 'text', 'image', 'video', 'document']] = Field(alias="headerType")
    header_text: Optional[str] = Field(alias="headerText")
    header_file: Optional[str] = Field(alias="headerFile")
    body_text: str = Field(alias="bodyText")
    footer_text: Optional[str] = Field(alias="footerText")
    buttons: Optional[List[dict]] = Field(alias="buttons")
    variables: Optional[List[WhatsAppVariable]] = Field(alias="variables")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class WhatsAppTemplateAPIDTO(BaseModel):
    id: str = Field(alias="id")
    template_id: str = Field(alias="template_id")
    name: str = Field(alias="name")
    label: str = Field(alias="label")
    description: str = Field(alias="description")
    alias: Optional[str] = Field(alias="alias", default=None)
    version: Optional[int] = Field(alias="version", default=None)
    category: str = Field(alias="category")
    segment: str = Field(alias="segment")
    periodicity: str = Field(alias="periodicity")
    client_id: str = Field(alias="clientId")
    language: Optional[str] = Field(alias="language", default=None)
    header_type: Optional[str] = Field(alias="headerType", default=None)
    header_text: Optional[str] = Field(alias="headerText", default=None)
    header_file: Optional[str] = Field(alias="headerFile", default=None)
    body_text: str = Field(alias="bodyText")
    footer_text: Optional[str] = Field(alias="footerText", default=None)
    buttons: Optional[List[dict]] = Field(alias="buttons", default=None)
    variables: Optional[List[WhatsAppVariable]] = Field(alias="variables", default=None)
    status: Optional[str] = Field(alias="status", default=None)
    tenant: str = Field(alias="tenant")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class WhatsAppTemplateUpdate(BaseModel):
    name: Optional[str] = Field(alias="name", default=None)
    label: Optional[str] = Field(alias="label", default=None)
    description: Optional[str] = Field(alias="description", default=None)
    category: Optional[str] = Field(alias="category", default=None)
    segment: Optional[str] = Field(alias="segment", default=None)
    periodicity: Optional[str] = Field(alias="periodicity", default=None)
    client_id: Optional[str] = Field(alias="clientId", default=None)
    language: Optional[str] = Field(alias="language", default=None)
    header_type: Optional[str] = Field(alias="headerType", default=None)
    header_text: Optional[str] = Field(alias="headerText", default=None)
    header_file: Optional[str] = Field(alias="headerFile", default=None)
    body_text: Optional[str] = Field(alias="bodyText", default=None)
    footer_text: Optional[str] = Field(alias="footerText", default=None)
    buttons: Optional[List[dict]] = Field(alias="buttons", default=None)
    variables: Optional[List[WhatsAppVariable]] = Field(alias="variables", default=None)

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class PublishRequest(BaseModel):
    connector_id: str = Field(alias="connectorId")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class SendRequest(BaseModel):
    connector_id: Optional[str] = Field(alias="connectorId")
    country_code: Optional[str] = Field(alias="countryCode", default=None)
    phone_number: str = Field(alias="phoneNumber")
    variables: Dict[str, Any] = Field(alias="variables")
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class WhatsAppTemplateSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/templates", header_builder, renew_token,
                         WhatsAppTemplateAPIDTO.parse_obj(data))


class WhatsAppTemplateAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/templates", header_builder, renew_token,
                         WhatsAppTemplateAPIDTO.parse_obj(data))


class WhatsAppTemplateSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, sync_resource=WhatsAppTemplateSync,
                         retrieve_data_model=WhatsAppTemplateAPIDTO,
                         create_data_model=WhatsAppTemplateRequest,
                         update_data_model=WhatsAppTemplateUpdate,
                         resource="/conversational/templates")

    @retry_on_401
    def publish_template(self, template_id: str, request_body: Dict[str, Any]):
        request_data = PublishRequest(
            **request_body
        )
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            request = client.post(
                f"/v1/conversational/templates/{template_id}/publish",
                json=request_data.dict(by_alias=True),
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(request)
            return request.json()

    @retry_on_401
    def send_template(self, template_id: str, request_body: Dict[str, Any]):
        request_data = SendRequest(
            **request_body
        )
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            request = client.post(
                f"/v1/conversational/templates/{template_id}/send",
                json=request_data.dict(by_alias=True),
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(request)


class WhatsAppTemplateAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, async_resource=WhatsAppTemplateAsync,
                         retrieve_data_model=WhatsAppTemplateAPIDTO,
                         create_data_model=WhatsAppTemplateRequest,
                         update_data_model=WhatsAppTemplateUpdate,
                         resource="/conversational/templates")

    @retry_on_401_async
    async def publish_template(self, template_id: str, request_body: Dict[str, Any]):
        request_data = PublishRequest(
            **request_body
        )
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                f"/v1/conversational/templates/{template_id}/publish",
                json=request_data.dict(by_alias=True),
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return response.json()

    @retry_on_401_async
    async def send_template(self, template_id: str, request_body: Dict[str, Any]):
        request_data = SendRequest(
            **request_body
        )
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                f"/v1/conversational/templates/{template_id}/send",
                json=request_data.dict(by_alias=True),
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)




import httpx
from pydantic import BaseModel, Field
from typing import Optional, Dict
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class CMSSettingsAPIDTO(BaseModel):
    id: str
    main_partner_id: Optional[str] = Field(alias='mainPartnerId', default=None)
    created_at: str = Field(alias='createdAt')
    updated_at: Optional[str] = Field(alias='updatedAt')

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateCMSSettingsDTO(BaseModel):
    main_partner_id: Optional[str] = Field(alias='mainPartnerId', default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CMSSettingsSync(GenericSyncResource):
    data: CMSSettingsAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "cms-settings", header_builder, renew_token, CMSSettingsAPIDTO.parse_obj(data))


class CMSSettingsAsync(GenericAsyncResource):
    data: CMSSettingsAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "cms-settings", header_builder, renew_token, CMSSettingsAPIDTO.parse_obj(data))


class CMSSettingsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=CMSSettingsSync,
                         retrieve_data_model=CMSSettingsAPIDTO,
                         create_data_model=None,
                         update_data_model=UpdateCMSSettingsDTO,
                         resource="application/cms-settings")

    @retry_on_401
    def retrieve(self):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/{self.resource}",
                headers=self.build_headers(),
                timeout=30
            )
            if response.status_code == 200:
                return self.sync_resource(
                    base_url=self.altscore_client._borrower_central_base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    data=self.retrieve_data_model.parse_obj(response.json())
                )
            elif response.status_code in [404]:
                return None

            raise_for_status_improved(response)

    @retry_on_401
    def put(self, patch_data: Dict) -> str:
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.put(
                f"/v1/{self.resource}",
                headers=self.build_headers(),
                json=self.update_data_model.parse_obj(patch_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            return self.retrieve_data_model.parse_obj(response.json())



class CMSSettingsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=CMSSettingsAsync,
                         retrieve_data_model=CMSSettingsAPIDTO,
                         create_data_model=None,
                         update_data_model=UpdateCMSSettingsDTO,
                         resource="application/cms-settings")


    @retry_on_401_async
    async def retrieve(self):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/{self.resource}",
                headers=self.build_headers(),
                timeout=30
            )
            if response.status_code == 200:
                return self.async_resource(
                    base_url=self.altscore_client._borrower_central_base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    data=self.retrieve_data_model.parse_obj(response.json())
                )
            elif response.status_code in [404]:
                return None

            raise_for_status_improved(response)


    @retry_on_401_async
    async def put(self, patch_data: Dict) -> str:
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.put(
                f"/v1/{self.resource}",
                headers=self.build_headers(),
                json=self.update_data_model.parse_obj(patch_data).dict(by_alias=True),
                timeout=30
            )
            raise_for_status_improved(response)
            return self.retrieve_data_model.parse_obj(response.json())



from pydantic import BaseModel, Field
from typing import List, Optional
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.cms.model.common import Money


class Reservation(BaseModel):
    id: Optional[str] = Field(alias="id", default=None)
    type: Optional[str] = Field(alias="type", default=None)
    amount: Money
    status: str
    created_at: str = Field(alias="createdAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class Percentages(BaseModel):
    available: float
    consumed: float

class CreditLine(BaseModel):
    assigned: Money
    available: Money
    consumed: Money
    reservations: Optional[List[Reservation]] = Field(alias="reservations", default=None)
    percentages: Optional[Percentages] = Field(alias="percentages", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class HistoryItem(BaseModel):
    createdAt: str
    amount: Money
    reason: Optional[str] = Field(alias="reason", default=None)
    userId: str

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class CreditAccountAPIDTO(BaseModel):
    client_id: str = Field(alias="clientId")
    partner_id: str = Field(alias="partnerId")
    borrower_id: str = Field(alias="borrowerId")
    external_id: str = Field(alias="externalId")
    product_family: str = Field(alias="productFamily")
    status: str = Field(alias="status")
    credit_line: CreditLine = Field(alias="creditLine")
    history: Optional[List[HistoryItem]] = Field(alias="history", default=None)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

    def is_enabled(self):
        return self.status == "enabled"


class CreditAccountAsync:
    data: CreditAccountAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: CreditAccountAPIDTO):
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    @retry_on_401_async
    async def update(self, amount: str, currency: str, reason: str) -> None:
        """
        Updates credit account with the given amount and currency, and the reason for the update.
        It will replace the data in the object with the response from the API.
        """
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.patch(
                f"/v2/clients/{self.data.client_id}/credit-accounts/{self.data.product_family}",
                json={
                    "assigned": {
                        "amount": amount,
                        "currency": currency,
                    },
                    "reason": reason,
                },
                headers=self._header_builder(partner_id=self.data.partner_id),
                timeout=30
            )
            raise_for_status_improved(response)
            self.data = CreditAccountAPIDTO.parse_obj(response.json())

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}"


class CreditAccountSync:
    data: CreditAccountAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: CreditAccountAPIDTO):
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    @retry_on_401
    def update(self, amount: str, currency: str, reason: str) -> None:
        """
        Updates credit account with the given amount and currency, and the reason for the update.
        It will replace the data in the object with the response from the API.
        """
        with httpx.Client(base_url=self.base_url) as client:
            response = client.patch(
                f"/v2/clients/{self.data.client_id}/credit-accounts/{self.data.product_family}",
                json={
                    "assigned": {
                        "amount": amount,
                        "currency": currency,
                    },
                    "reason": reason,
                },
                headers=self._header_builder(partner_id=self.data.partner_id),
                timeout=30
            )
            raise_for_status_improved(response)
            self.data = CreditAccountAPIDTO.parse_obj(response.json())

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}"



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field, root_validator
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class Money(BaseModel):
    amount: str = Field(alias="amount")
    currency: str = Field(alias="currency")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

# DTO for field value history
class HistoricValueDTO(BaseModel):
    """Data transfer object for a historic field value"""
    reference_id: str = Field(alias="referenceId")
    value: Any = Field(alias="value")
    updated_at: str = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


# DTO for priced array items
class MoneyArrayItemDTO(BaseModel):
    """Data transfer object for a money array item"""
    key: str = Field(alias="key")
    label: Optional[str] = Field(alias="label", default=None)
    value: Money = Field(alias="value")

# DTO for asset fields
class AssetFieldDTO(BaseModel):
    """Data transfer object for asset fields"""
    id: str = Field(alias="id")
    asset_id: str = Field(alias="assetId")
    key: str = Field(alias="key")
    label: Optional[str] = Field(alias="label", default=None)
    value: Any = Field(alias="value")
    data_type: str = Field(alias="dataType")
    history: List[HistoricValueDTO] = Field(alias="history", default=[])
    tags: List[str] = Field(alias="tags", default=[])
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

    @root_validator(pre=False)
    def parse_history_values(cls, values):
        """Parse history values based on the field's data_type"""
        data_type = values.get("data_type")
        history = values.get("history", [])
        
        if not history:
            return values
        
        # Parse current value based on data_type
        current_value = values.get("value")
        if data_type == "money" and isinstance(current_value, dict):
            values["value"] = Money.parse_obj(current_value)
        elif data_type == "money_array" and isinstance(current_value, list):
            values["value"] = [MoneyArrayItemDTO.parse_obj(item) for item in current_value]
        
        # Parse history values based on data_type
        for hist_item in history:
            if data_type == "money" and isinstance(hist_item.value, dict):
                hist_item.value = Money.parse_obj(hist_item.value)
            elif data_type == "money_array" and isinstance(hist_item.value, list):
                hist_item.value = [MoneyArrayItemDTO.parse_obj(item) for item in hist_item.value]
        
        return values


class CreateAssetFieldRequest(BaseModel):
    """Model for creating a new asset field"""
    asset_id: str = Field(alias="assetId")
    key: str = Field(alias="key")
    value: Any = Field(alias="value")
    data_type: Optional[str] = Field(alias="dataType", default=None)
    reference_id: Optional[str] = Field(alias="referenceId", default=None)
    tags: List[str] = Field(alias="tags", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateAssetFieldRequest(BaseModel):
    """Model for updating an asset field"""
    asset_id: str = Field(alias="assetId")
    value: Any = Field(alias="value")
    data_type: Optional[str] = Field(alias="dataType", default=None)
    reference_id: Optional[str] = Field(alias="referenceId", default=None)  # Source of the value for history tracking
    tags: Optional[List[str]] = Field(alias="tags", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


# Resource classes for asset fields
class AssetFieldSync(GenericSyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "asset-fields", header_builder, renew_token, AssetFieldDTO.parse_obj(data))


class AssetFieldAsync(GenericAsyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "asset-fields", header_builder, renew_token, AssetFieldDTO.parse_obj(data))


# Module for asset fields - synchronous
class AssetFieldsSyncModule(GenericSyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=AssetFieldSync,
                         retrieve_data_model=AssetFieldDTO,
                         create_data_model=CreateAssetFieldRequest,
                         update_data_model=UpdateAssetFieldRequest,
                         resource="asset-fields")

    @retry_on_401
    def get_by_asset_id(self, asset_id: str, page: int = 1, per_page: int = 100):
        """
        Get asset fields by asset ID
        
        Args:
            asset_id: The ID of the asset
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[AssetFieldDTO]: List of asset fields
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/asset-fields",
                params={
                    "asset-id": asset_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [AssetFieldDTO.parse_obj(data) for data in response.json()]

    @retry_on_401
    def get_by_key(self, asset_id: str, key: str):
        """
        Get an asset field by its key

        Args:
            asset_id: The ID of the asset
            key: The field key

        Returns:
            List[AssetFieldDTO]: List of asset fields matching the key
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/asset-fields",
                params={
                    "key": key,
                    "asset-id": asset_id,
                    "per-page": 100,
                    "page": 1
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [AssetFieldDTO.parse_obj(data) for data in response.json()]


# Module for asset fields - asynchronous
class AssetFieldsAsyncModule(GenericAsyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=AssetFieldAsync,
                         retrieve_data_model=AssetFieldDTO,
                         create_data_model=CreateAssetFieldRequest,
                         update_data_model=UpdateAssetFieldRequest,
                         resource="asset-fields")

    @retry_on_401_async
    async def get_by_asset_id(self, asset_id: str, page: int = 1, per_page: int = 100):
        """
        Get asset fields by asset ID
        
        Args:
            asset_id: The ID of the asset
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[AssetFieldDTO]: List of asset fields
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/asset-fields",
                params={
                    "asset-id": asset_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [AssetFieldDTO.parse_obj(data) for data in response.json()]

    @retry_on_401_async
    async def get_by_key(self, asset_id: str, key: str):
        """
        Get an asset field by its key
        
        Args:
            asset_id: The ID of the asset
            key: The field key
            
        Returns:
            List[AssetFieldDTO]: List of asset fields matching the key
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/asset-fields",
                params={
                    "key": key,
                    "asset-id": asset_id,
                    "per-page": 100,
                    "page": 1
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [AssetFieldDTO.parse_obj(data) for data in response.json()]



import datetime as dt
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
import httpx
from altscore.altdata.helpers import build_headers
from pydantic import BaseModel, validator, Field
from altscore.altdata.model.common_schemas import SourceConfig
from altscore.altdata.utils.dataframes import df_to_base64
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from dateutil.parser import parse
import json


class BatchFileRequest(BaseModel):
    name: str
    base64: str


class BatchDataRequest(BaseModel):
    label: str = Field(alias="label")
    sources_config: List[SourceConfig] = Field(alias="sourcesConfig")
    requests: List[Dict] = Field(alias="requests", default=[])
    file: Optional[BatchFileRequest] = Field(alias="file", default=None)


class BatchSourceSuccessStats(BaseModel):
    source_id: str = Field(alias="sourceId")
    total_count: int = Field(alias="totalCount")
    success_count: int = Field(alias="successCount")
    failed_count: int = Field(alias="failedCount")
    processing_count: int = Field(alias="processingCount")
    non_retryable_count: int = Field(alias="nonRetryableCount")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BatchStatus(BaseModel):
    batch_id: str = Field(alias="batchId")
    requested_at: Union[dt.datetime, str] = Field(alias="requestedAt")
    request_count: int = Field(alias="requestCount")
    status: str = Field(alias="status")
    progress_pct: float = Field(alias="progressPct")
    success_pct: float = Field(alias="successPct")
    failed_pct: float = Field(alias="failedPct")
    non_retryable_pct: float = Field(alias="nonRetryablePct")
    source_stats: List[BatchSourceSuccessStats] = Field(alias="sourceStats")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

    @validator("requested_at")
    def parse_requested_at(cls, v):
        if isinstance(v, str):
            return parse(v, yearfirst=True)
        return v

    def __str__(self):
        r = self.dict(by_alias=True)
        if isinstance(r.get("requestedAt"), dt.datetime):
            r["requestedAt"] = r["requestedAt"].isoformat()
        return json.dumps(r, indent=4, ensure_ascii=False)

    def print_source_stats(self):
        import pandas as pd
        df = pd.DataFrame([e.dict() for e in self.source_stats])
        print(df.to_markdown())


class BatchSyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    def new_batch_from_dataframe(self, df, label: str,
                                 sources_config: List[SourceConfig]):
        with httpx.Client(base_url=self.altscore_client._altdata_base_url) as client:
            payload = df_to_batch_payload(df=df, label=label, sources_config=sources_config)
            batch_response = client.post(
                "/v1/batches",
                json=payload,
                headers=self.build_headers(),
                timeout=500
            )
            raise_for_status_improved(batch_response)
            batch_id = batch_response.json()["batchId"]
            return BatchSync(
                base_url=self.altscore_client._altdata_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=BatchData(batch_id=batch_id, label=label, sources_config=sources_config)
            )

    def retrieve(self, batch_id: str):
        with httpx.Client(base_url=self.altscore_client._altdata_base_url) as client:
            response = client.get(
                f"/v1/batches/{batch_id}",
                headers=self.build_headers(),
                timeout=500
            )
            raise_for_status_improved(response)
            return BatchSync(
                base_url=self.altscore_client._altdata_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=BatchData(
                    batch_id=batch_id,
                    label=response.json()["label"],
                    sources_config=response.json()["sourcesConfig"]
                )
            )


class BatchAsyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    async def new_batch_from_dataframe(self, df, label: str, sources_config: List[SourceConfig]):
        async with httpx.AsyncClient(base_url=self.altscore_client._altdata_base_url) as client:
            payload = df_to_batch_payload(df=df, label=label, sources_config=sources_config)
            batch_response = await client.post(
                "/v1/batches",
                json=payload,
                headers=self.build_headers(),
                timeout=500
            )
            raise_for_status_improved(batch_response)
            batch_id = batch_response.json()["batchId"]
            return BatchSync(
                base_url=self.altscore_client._altdata_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=BatchData(batch_id=batch_id, label=label, sources_config=sources_config)
            )

    async def retrieve(self, batch_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._altdata_base_url) as client:
            response = await client.get(
                f"/v1/batches/{batch_id}",
                headers=self.build_headers(),
                timeout=500
            )
            raise_for_status_improved(response)
            return BatchSync(
                base_url=self.altscore_client._altdata_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=BatchData(
                    batch_id=batch_id,
                    label=response.json()["label"],
                    sources_config=response.json()["sourcesConfig"]
                )
            )


class BatchBase:

    def __init__(self, base_url):
        self.base_url = base_url

    def _status(self, batch_id: str):
        return f"{self.base_url}/v1/batches/{batch_id}/status"

    def _retry(self, batch_id: str):
        return f"{self.base_url}/v1/batches/{batch_id}/retry"

    def _export(self, batch_id: str):
        return f"{self.base_url}/v1/batches/{batch_id}/export"


class BatchData:

    def __init__(self, batch_id, label, sources_config):
        self.batch_id = batch_id
        self.label = label
        self.sources_config = sources_config
        self.status: Optional[BatchStatus] = None
        self.export_urls: Optional[Dict[str, str]] = None


class BatchSync(BatchBase):

    def __init__(self, base_url, header_builder, renew_token, data: BatchData):
        super().__init__(base_url)
        self.header_builder = header_builder
        self.renew_token = renew_token
        self.data: BatchData = data

    @property
    def batch_id(self):
        return self.data.batch_id

    @property
    def label(self):
        return self.data.label

    @property
    def status(self):
        return self.data.status

    @property
    def export_urls(self):
        return self.data.export_urls

    @retry_on_401
    def get_status(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.get(self._status(batch_id=self.data.batch_id),
                                  headers=self.header_builder())
            raise_for_status_improved(response)
            self.data.status = BatchStatus.parse_obj(response.json())

    @retry_on_401
    def retry(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(self._retry(batch_id=self.data.batch_id),
                                   headers=self.header_builder())
            raise_for_status_improved(response)

    @retry_on_401
    def _get_export_urls(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                self._export(batch_id=self.data.batch_id),
                headers=self.header_builder(),
                timeout=500
            )
            raise_for_status_improved(response)
            data = response.json()
            assert data.get("dataExportUrl") is not None and data.get(
                "dataExportUrl") != "", "Failed to parse response, contact support or update SDK"
            assert data.get("sourceDataExportUrl") is not None and data.get(
                "sourceDataExportUrl") != "", "Failed to parse response, contact support or update SDK"
            self.data.export_urls = data

    def export_to_dataframe(self):
        import pandas as pd
        if self.export_urls is None:
            self._get_export_urls()
        return pd.read_csv(
            self.export_urls["dataExportUrl"], encoding="utf-8",
            dtype={"personId": str, "taxId": str, "foreignKey": str, "email": str, "phone": str}
        )

    def export_source_data_to_dict(self):
        if self.export_urls is None:
            self._get_export_urls()
        with httpx.Client() as client:
            req = client.get(self.export_urls["sourceDataExportUrl"], timeout=500)
            data = [json.loads(e) for e in req.content.decode("utf8").split("\n") if len(e) > 0]
            return data


class BatchAsync(BatchBase):

    def __init__(self, base_url, header_builder, renew_token, data: BatchData):
        super().__init__(base_url)
        self.header_builder = header_builder
        self.renew_token = renew_token
        self.data: BatchData = data

    @property
    def batch_id(self):
        return self.data.batch_id

    @property
    def label(self):
        return self.data.label

    @property
    def status(self):
        return self.data.status

    @property
    def export_urls(self):
        return self.data.export_urls

    @retry_on_401_async
    async def get_status(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.get(self._status(batch_id=self.data.batch_id),
                                        headers=self.header_builder())
            raise_for_status_improved(response)
            self.data.status = BatchStatus.parse_obj(response.json())

    @retry_on_401_async
    async def retry(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(self._retry(batch_id=self.data.batch_id),
                                         headers=self.header_builder())
            raise_for_status_improved(response)

    async def _get_export_urls(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                self._export(batch_id=self.data.batch_id),
                headers=self.header_builder(),
                timeout=500
            )
            raise_for_status_improved(response)
            data = response.json()
            assert data.get("dataExportUrl") is not None and data.get(
                "dataExportUrl") != "", "Failed to parse response, contact support or update SDK"
            assert data.get("sourceDataExportUrl") is not None and data.get(
                "sourceDataExportUrl") != "", "Failed to parse response, contact support or update SDK"
            self.data.export_urls = data

    async def export_to_dataframe(self):
        import pandas as pd
        if self.export_urls is None:
            await self._get_export_urls()
        return pd.read_csv(
            self.export_urls["dataExportUrl"], encoding="utf-8",
            dtype={"personId": str, "taxId": str, "foreignKey": str, "email": str, "phone": str}
        )

    async def export_source_data_to_dict(self):
        if self.export_urls is None:
            await self._get_export_urls()

        async with httpx.AsyncClient() as client:
            req = await client.get(self.export_urls["sourceDataExportUrl"], timeout=500)
            data = [json.loads(e) for e in req.content.decode("utf8").split("\n") if len(e) > 0]
            return data


def df_to_batch_payload(df, label, sources_config):
    import numpy as np
    data = df.replace({np.nan: None})
    base_64 = df_to_base64(data)
    sources_config_obj = [SourceConfig.parse_obj(x) for x in sources_config]
    payload = BatchDataRequest(
        label=label,
        sourcesConfig=sources_config_obj,
        file=BatchFileRequest(name="df_from_sdk.csv", base64=base_64),
    ).dict(by_alias=True)
    return payload



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule

class ListStatus:
    PENDING = "pending"
    APPLIED = "applied"
    NO_HIT = "no_hit"

class SimilarEntity(BaseModel):
    entity_type: str = Field(alias="entityType")
    key: str = Field(alias="key")
    proposed_value: Any = Field(alias="proposedValue")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class Similar(BaseModel):
    label: str = Field(alias="label")
    description: Optional[str] = Field(alias="description")
    entities: List[SimilarEntity] = Field(alias="entities")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ListOfSimilarAPIDTO(BaseModel):
    id: str = Field(alias="id")
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    deal_id: Optional[str] = Field(alias="dealId", default=None)
    execution_id: Optional[str] = Field(alias="executionId")
    list_of_similar: List[Similar] = Field(alias="listOfSimilar")
    status: Optional[str] = Field(alias="status")
    applied_by: Optional[str] = Field(alias="appliedBy")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateListOfSimilar(BaseModel):
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    deal_id: Optional[str] = Field(alias="dealId", default=None)
    execution_id: Optional[str] = Field(alias="executionId")
    list_of_similar: List[Similar] = Field(alias="listOfSimilar")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ListOfSimilarSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "list-of-similar", header_builder, renew_token, ListOfSimilarAPIDTO.parse_obj(data))

    @retry_on_401
    def apply(self, index: int, retry_workflow: bool = False):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                f"/v1/list-of-similar/{self.data.id}/apply",
                headers=self._header_builder(),
                timeout=300,
                json={
                    "index": index,
                    "retryWorkflow": retry_workflow
                }
            )
            raise_for_status_improved(response)

    @retry_on_401
    def report_no_hit(self, retry_workflow: bool = False):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                f"/v1/list-of-similar/{self.data.id}/no-hit",
                headers=self._header_builder(),
                timeout=300,
                json={
                    "retryWorkflow": retry_workflow
                }
            )
            raise_for_status_improved(response)


class ListOfSimilarAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "list-of-similar", header_builder, renew_token, ListOfSimilarAPIDTO.parse_obj(data))

    @retry_on_401_async
    async def apply(self, index: int, retry_workflow: bool = False):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                f"/v1/list-of-similar/{self.data.id}/apply",
                headers=self._header_builder(),
                timeout=300,
                json={
                    "index": index,
                    "retryWorkflow": retry_workflow
                }
            )
            raise_for_status_improved(response)

    @retry_on_401_async
    async def report_no_hit(self, retry_workflow: bool = False):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                f"/v1/list-of-similar/{self.data.id}/no-hit",
                headers=self._header_builder(),
                timeout=300,
                json={
                    "retryWorkflow": retry_workflow
                }
            )
            raise_for_status_improved(response)


class ListOfSimilarSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=ListOfSimilarSync,
                         retrieve_data_model=ListOfSimilarAPIDTO,
                         create_data_model=CreateListOfSimilar,
                         update_data_model=None,
                         resource="list-of-similar")


class ListOfSimilarAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=ListOfSimilarAsync,
                         retrieve_data_model=ListOfSimilarAPIDTO,
                         create_data_model=CreateListOfSimilar,
                         update_data_model=None,
                         resource="list-of-similar")



from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async


class RuleAlert(BaseModel):
    level: int = Field(alias="level")
    message: str = Field(alias="message")


class RuleAPIDTO(BaseModel):
    id: str = Field(alias="id")
    code: str = Field(alias="code")
    label: str = Field(alias="label")
    alerts: List[RuleAlert] = Field(alias="alerts", default=[])
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

    def get_alert_by_level(self, level: int):
        for alert in self.alerts:
            if alert.level == level:
                return alert
        return None


class CreateRuleDTO(BaseModel):
    label: str = Field(alias="label")
    code: str = Field(alias="code")
    alerts: List[RuleAlert] = Field(alias="alerts")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class RulesSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "rules", header_builder, renew_token, RuleAPIDTO.parse_obj(data))


class RulesAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "rules", header_builder, renew_token, RuleAPIDTO.parse_obj(data))


class RulesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=RulesSync,
                         retrieve_data_model=RuleAPIDTO,
                         create_data_model=CreateRuleDTO,
                         update_data_model=CreateRuleDTO,
                         resource="rules")

    @retry_on_401
    def retrieve_by_code(self, code: str):
        query_params = {
            "code": code
        }
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/{self.resource}",
                headers=self.build_headers(),
                params=query_params,
                timeout=30
            )
            raise_for_status_improved(response)
            res = [self.sync_resource(
                base_url=self.altscore_client._borrower_central_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=self.retrieve_data_model.parse_obj(e)
            ) for e in response.json()]

            if len(res) == 0:
                return None
            return res[0]


class RulesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=RulesAsync,
                         retrieve_data_model=RuleAPIDTO,
                         create_data_model=CreateRuleDTO,
                         update_data_model=CreateRuleDTO,
                         resource="rules")

    @retry_on_401_async
    async def retrieve_by_code(self, code: str):
        query_params = {
            "code": code
        }
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/{self.resource}",
                headers=self.build_headers(),
                params=query_params,
                timeout=30
            )
            raise_for_status_improved(response)
            res = [self.async_resource(
                base_url=self.altscore_client._borrower_central_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=self.retrieve_data_model.parse_obj(e)
            ) for e in response.json()]

            if len(res) == 0:
                return None
            return res[0]



from pydantic import BaseModel, Field
from typing import List, Optional, Dict, Union
from altscore.altdata.helpers import build_headers
from altscore.altdata.model.common_schemas import SourceConfig
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
import json
import httpx


class Address(BaseModel):
    street1: str = Field(alias="street1")
    street2: Optional[str] = Field(alias="street2", default=None)
    neighborhood: str = Field(alias="neighborhood", default=None)
    district: str = Field(alias="district", default=None)
    city: str = Field(alias="city", default=None)
    state: str = Field(alias="state", default=None)
    province: Optional[str] = Field(alias="province", default=None)
    zip_code: str = Field(alias="zipCode", default=None)
    country: str = Field(alias="country", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Location(BaseModel):
    lat: float = Field(alias="lat")
    lon: float = Field(alias="lon")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class InputKeys(BaseModel):
    foreign_key: Optional[str] = Field(alias="foreignKey", default=None)
    person_id: Optional[str] = Field(alias="personId", default=None)
    person_id_expedition_date: Optional[str] = Field(alias="personIdExpeditionDate", default=None)
    name: Optional[str] = Field(alias="name", default=None)
    first_name: Optional[str] = Field(alias="firstName", default=None)
    middle_name: Optional[str] = Field(alias="middleName", default=None)
    paternal_surname: Optional[str] = Field(alias="paternalSurname", default=None)
    maternal_surname: Optional[str] = Field(alias="maternalSurname", default=None)
    birth_date: Optional[str] = Field(alias="birthDate", default=None)
    nationality: Optional[str] = Field(alias="nationality", default=None)
    phone: Optional[str] = Field(alias="phone", default=None)
    email: Optional[str] = Field(alias="email", default=None)
    vehicle_license_plate: Optional[str] = Field(alias="vehicleLicensePlate", default=None)
    address: Optional[Address] = Field(alias="address", default=None)
    location: Optional[Location] = Field(alias="location", default=None)
    tax_id: Optional[str] = Field(alias="taxId", default=None)
    legal_name: Optional[str] = Field(alias="legalName", default=None)
    dba: Optional[str] = Field(alias="dba", default=None)
    business_id: Optional[str] = Field(alias="businessId", default=None)
    legal_rep_id: Optional[str] = Field(alias="legalRepId", default=None)
    country_code: Optional[str] = Field(alias="countryCode", default=None)
    date_to_analyze: Optional[str] = Field(alias="dateToAnalyze", default=None)
    authorization_reference: Optional[str] = Field(alias="authorizationReference", default=None)
    passport_number: Optional[str] = Field(alias="passportNumber", default=None)
    items: Optional[List] = Field(alias="items", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class SourceCallSummary(BaseModel):
    source_id: str = Field(alias="sourceId")
    version: str = Field(alias="version")
    status: Optional[str] = Field(alias="status", default=None)
    is_success: bool = Field(alias="isSuccess")
    error_message: Optional[str] = Field(alias="errorMessage", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

    def __repr__(self):
        return f"<SourceCallSummary: {self.source_id}_{self.version}, isSuccess:{self.is_success}>"


class StatusAPIDTO(BaseModel):
    request_id: str = Field(alias="requestId")
    status: str = Field(alias="status")
    requested_at: str = Field(alias="requestedAt")
    call_summary: List[SourceCallSummary] = Field(alias="callSummary")


class RequestStatus:
    def __init__(self, data: StatusAPIDTO):
        self.data = data

    @classmethod
    def from_api(cls, response: Dict):
        return cls(data=StatusAPIDTO.parse_obj(response))

    def are_all_source_calls_success(self):
        return all([s.is_success for s in self.data.call_summary])

    @property
    def status(self):
        return self.data.status

    def is_complete(self):
        return self.status == "complete"


class RequestAPIDTO(BaseModel):
    request_id: str = Field(alias="requestId")
    flat_data: Dict = Field(alias="data")
    call_summary: Optional[List[SourceCallSummary]] = Field(alias="callSummary", default=[])
    source_data: Dict = Field(alias="sourceData", default=None)
    inputs: Dict
    requested_at: str = Field(alias="requestedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class RequestResult:
    def __init__(self, data: RequestAPIDTO):
        self.data = data

    def __repr__(self):
        r = self.data.dict(by_alias=True)
        return json.dumps(r, indent=4, ensure_ascii=False)

    @classmethod
    def from_api(cls, response: Dict):
        return cls(data=RequestAPIDTO.parse_obj(response))

    @property
    def call_summary(self):
        return self.data.call_summary

    def are_all_source_calls_success(self):
        return all([s.is_success for s in self.data.call_summary])

    def get_source_call_summary(self, source_id: str):
        return next((s for s in self.data.call_summary if s.source_id == source_id), None)

    def get_data(self, source_id: str):
        return self.data.flat_data.get(source_id, None)

    def get_source_data(self, source_id: str):
        return self.data.source_data.get(source_id, None)

    def to_package(self, source_id: str):
        source_call_summary = self.get_source_call_summary(source_id).dict(by_alias=True)
        return {
            "sourceId": source_id,
            "version": source_call_summary["version"],
            "isSuccess": source_call_summary["isSuccess"],
            "requestId": self.data.request_id,
            "data": self.get_data(source_id),
            "sourceData": self.get_source_data(source_id),
            "inputs": self.data.inputs,
            "requestedAt": self.data.requested_at,
        }


class RequestSyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401
    def new_sync(self, input_keys: Union[InputKeys, Dict], sources_config: List[Union[Dict, SourceConfig]],
                 timeout: Optional[int] = None, execution_id: Optional[str] = None, workflow_id: Optional[str] = None,
                 batch_id: Optional[str] = None):
        if isinstance(input_keys, dict):
            # to validate the input keys
            input_keys = InputKeys.parse_obj(input_keys)
        payload = input_keys.dict(by_alias=True, exclude_none=True)
        # to validate the sources config model
        sources_config = [SourceConfig.parse_obj(s) if isinstance(s, dict) else s for s in sources_config]

        if timeout is not None:
            payload["timeout"] = timeout

        if execution_id is not None:
            payload["executionId"] = execution_id

        if workflow_id is not None:
            payload["workflowId"] = workflow_id

        if batch_id is not None:
            payload["batchId"] = batch_id

        payload["sourcesConfig"] = [s.dict(by_alias=True) for s in sources_config]
        with httpx.Client(base_url=self.altscore_client._altdata_base_url) as client:
            r = client.post(
                url="/v1/requests/sync",
                json=payload,
                headers=self.build_headers(),
                timeout=500
                # dont confuse with the timeout in the payload, this is the timeout for the request, not the sources
            )
            raise_for_status_improved(r)
            sync_data_response = r.json()
            return RequestResult.from_api(sync_data_response)

    @retry_on_401
    def new_async(self, input_keys: Union[InputKeys, Dict], sources_config: List[Union[Dict, SourceConfig]],
                  execution_id: Optional[str] = None, workflow_id: Optional[str] = None,
                  batch_id: Optional[str] = None):
        if isinstance(input_keys, dict):
            # to validate the input keys
            input_keys = InputKeys.parse_obj(input_keys)
        payload = input_keys.dict(by_alias=True, exclude_none=True)
        # to validate the sources config model
        sources_config = [SourceConfig.parse_obj(s) if isinstance(s, dict) else s for s in sources_config]

        if execution_id is not None:
            payload["executionId"] = execution_id

        if workflow_id is not None:
            payload["workflowId"] = workflow_id

        if batch_id is not None:
            payload["batchId"] = batch_id

        payload["sourcesConfig"] = [s.dict(by_alias=True) for s in sources_config]
        with httpx.Client(base_url=self.altscore_client._altdata_base_url) as client:
            r = client.post(
                url="/v1/requests/async",
                json=payload,
                headers=self.build_headers(),
                timeout=500
            )
            raise_for_status_improved(r)
            sync_data_response = r.json()
            return AsyncRequestSync(
                base_url=self.altscore_client._altdata_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                request_id=sync_data_response["requestId"]
            )

    def retrieve(self, request_id: str):
        return AsyncRequestSync(
            base_url=self.altscore_client._altdata_base_url,
            header_builder=self.build_headers,
            renew_token=self.renew_token,
            request_id=request_id
        )


class RequestAsyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401_async
    async def new_sync(self, input_keys: Union[InputKeys, Dict], sources_config: List[Union[Dict, SourceConfig]],
                       timeout: Optional[int] = None, execution_id: Optional[str] = None,
                       workflow_id: Optional[str] = None, batch_id: Optional[str] = None):
        if isinstance(input_keys, dict):
            # to validate the input keys
            input_keys = InputKeys.parse_obj(input_keys)
        payload = input_keys.dict(by_alias=True, exclude_none=True)
        # to validate the sources config model
        sources_config = [SourceConfig.parse_obj(s) if isinstance(s, dict) else s for s in sources_config]
        if timeout is not None:
            payload["timeout"] = timeout

        if execution_id is not None:
            payload["executionId"] = execution_id

        if workflow_id is not None:
            payload["workflowId"] = workflow_id

        if batch_id is not None:
            payload["batchId"] = batch_id

        payload["sourcesConfig"] = [s.dict(by_alias=True) for s in sources_config]
        async with httpx.AsyncClient(base_url=self.altscore_client._altdata_base_url) as client:
            r = await client.post(
                url="/v1/requests/sync",
                json=payload,
                headers=self.build_headers(),
                timeout=500
            )
            raise_for_status_improved(r)
            sync_data_response = r.json()
            return RequestResult.from_api(sync_data_response)

    @retry_on_401_async
    async def new_async(self, input_keys: Union[InputKeys, Dict], sources_config: List[Union[Dict, SourceConfig]],
                        execution_id: Optional[str] = None, workflow_id: Optional[str] = None,
                        batch_id: Optional[str] = None):
        if isinstance(input_keys, dict):
            # to validate the input keys
            input_keys = InputKeys.parse_obj(input_keys)
        payload = input_keys.dict(by_alias=True, exclude_none=True)
        # to validate the sources config model
        sources_config = [SourceConfig.parse_obj(s) if isinstance(s, dict) else s for s in sources_config]

        if execution_id is not None:
            payload["executionId"] = execution_id

        if workflow_id is not None:
            payload["workflowId"] = workflow_id

        if batch_id is not None:
            payload["batchId"] = batch_id

        payload["sourcesConfig"] = [s.dict(by_alias=True) for s in sources_config]
        async with httpx.AsyncClient(base_url=self.altscore_client._altdata_base_url) as client:
            r = await client.post(
                url="/v1/requests/async",
                json=payload,
                headers=self.build_headers(),
                timeout=500
            )
            raise_for_status_improved(r)
            sync_data_response = r.json()
            return AsyncRequestAsync(
                base_url=self.altscore_client._altdata_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                request_id=sync_data_response["requestId"]
            )

    def retrieve(self, request_id: str):
        return AsyncRequestSync(
            base_url=self.altscore_client._altdata_base_url,
            header_builder=self.build_headers,
            renew_token=self.renew_token,
            request_id=request_id
        )


class RequestsBase:

    def __init__(self, base_url):
        self.base_url = base_url

    def _get(self, request_id: str):
        return f"{self.base_url}/v1/requests/{request_id}"

    def _get_status(self, request_id: str):
        return f"{self.base_url}/v1/requests/{request_id}/status"

    def _retry(self, request_id: str):
        return f"{self.base_url}/v1/requests/{request_id}/retry"


class AsyncRequestSync(RequestsBase):

    def __init__(self, base_url, header_builder, renew_token, request_id: str):
        super().__init__(base_url)
        self.base_url = base_url
        self.header_builder = header_builder
        self.renew_token = renew_token
        self.request_id = request_id

    @property
    def id(self):
        return self.request_id

    @retry_on_401
    def pull(self):
        with httpx.Client(base_url=self.base_url) as client:
            r = client.get(
                url=self._get(self.id),
                headers=self.header_builder()
            )
            raise_for_status_improved(r)
            return RequestResult.from_api(r.json())

    @retry_on_401
    def get_status(self):
        with httpx.Client(base_url=self.base_url) as client:
            r = client.get(
                url=self._get_status(self.id),
                headers=self.header_builder()
            )
            raise_for_status_improved(r)
            return RequestStatus.from_api(r.json())


class AsyncRequestAsync(RequestsBase):

    def __init__(self, base_url, header_builder, renew_token, request_id: str):
        super().__init__(base_url)
        self.base_url = base_url
        self.header_builder = header_builder
        self.renew_token = renew_token
        self.request_id = request_id

    @property
    def id(self):
        return self.request_id

    @retry_on_401_async
    async def pull(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            r = await client.get(
                url=self._get(self.id),
                headers=self.header_builder()
            )
            raise_for_status_improved(r)
            return RequestResult.from_api(r.json())

    @retry_on_401_async
    async def get_status(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            r = await client.get(
                url=self._get_status(self.id),
                headers=self.header_builder()
            )
            raise_for_status_improved(r)
            return RequestStatus.from_api(r.json())



from pydantic import BaseModel, Field
from typing import Optional, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from altscore.borrower_central.model.accounting_amount import Amount


class LineItemAPIDTO(BaseModel):
    id: str = Field(alias="id")
    item_id: str = Field(alias="itemId")
    line_number_in_source: Optional[int] = Field(default=None, alias="lineNumberInSource")
    accounting_document_id: str = Field(alias="accountingDocumentId")
    item_code: Optional[str] = Field(default=None, alias="itemCode",
                                     description="Item code, if the item is not an stockable with sku")
    sku: Optional[str] = Field(default=None, alias="sku", description="Product / material code")
    description: Optional[str] = Field(default=None, alias="description")
    qty_raw: Optional[str] = Field(default=None, alias="qtyRaw", description="Quantity involved")
    qty_normalized: Optional[float] = Field(default=None, alias="qtyNormalized", description="Quantity involved")
    qty_unit: Optional[str] = Field(default=None, alias="qtyUnit", description="Unit of measure for quantity")
    unit_price: Optional[str] = Field(default=None, alias="unitPrice")
    unit_price_currency: Optional[str] = Field(default=None, alias="unitPriceCurrency")
    amount: Optional[Amount] = Field(default=None, alias="amount", description="amount")
    parsed_raw: Optional[Dict[str, Any]] = Field(default=None, alias="parsedRaw")
    metadata: Optional[Dict[str, Any]] = Field(default=None, alias="metadata")
    # Variance
    is_variance: Optional[bool] = Field(default=None, alias="isVariance")
    variance_kind: Optional[str] = Field(
        default=None,
        alias="varianceKind",
        description="PRICE, QUANTITY, BOTH"
    )
    variance_reason: Optional[str] = Field(
        default=None, alias="varianceReason",
        description="Reason for variance"
    )
    reference_unit_price: Optional[str] = Field(
        default=None, alias="referenceUnitPrice",
        description="Price from PO / contract (for PRICE variance)"
    )
    reference_amount: Optional[Amount] = Field(
        default=None, alias="referenceAmount",
        description="Reference amount from PO or internal"
    )
    reference_qty_raw: Optional[str] = Field(
        default=None, alias="referenceQtyRaw",
        description="Expected quantity (for QUANTITY variance)"
    )
    reference_qty_normalized: Optional[float] = Field(
        default=None, alias="referenceQtyNormalized",
        description="Expected quantity (for QUANTITY variance)"
    )
    # Return
    is_return: Optional[bool] = Field(default=None, alias="isReturn")
    reason_code: Optional[str] = Field(
        default=None, alias="reasonCode",
        description="Damaged, expired, decomiso, etc."
    )
    # Rebate
    is_rebate: Optional[bool] = Field(default=None, alias="isRebate")
    agreement_id: Optional[str] = Field(
        default=None, alias="agreementId",
        description="ID of trade-promotion / rebate contract"
    )
    period_start: Optional[str] = Field(
        default=None, alias="periodStart",
        description="Start of accrual period (yyyy-mm-dd)"
    )
    period_end: Optional[str] = Field(
        default=None, alias="periodEnd",
        description="End of accrual period (yyyy-mm-dd)"
    )
    # Parsing confidence
    confidence_score: Optional[float] = Field(
        default=None, alias="confidenceScore",
        description="Confidence score for parsing"
    )
    # Created and updated at
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(default=None, alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateLineItemDTO(BaseModel):
    accounting_document_id: str = Field(alias="accountingDocumentId")
    line_number_in_source: Optional[int] = Field(default=None, alias="lineNumberInSource")
    description: Optional[str] = Field(default=None, alias="description")
    qty_raw: Optional[str] = Field(default=None, alias="qtyRaw")
    qty_normalized: Optional[float] = Field(default=None, alias="qtyNormalized")
    qty_unit: Optional[str] = Field(default=None, alias="qtyUnit")
    unit_price: Optional[str] = Field(default=None, alias="unitPrice")
    unit_price_currency: Optional[str] = Field(default=None, alias="unitPriceCurrency")
    amount: Optional[Amount] = Field(default=None, alias="amount")
    sku: Optional[str] = Field(default=None, alias="sku")
    item_code: Optional[str] = Field(default=None, alias="itemCode")
    parsed_raw: Optional[Dict[str, Any]] = Field(default=None, alias="parsedRaw")
    metadata: Optional[Dict[str, Any]] = Field(default=None, alias="metadata")
    # Variance fields
    is_variance: Optional[bool] = Field(default=None, alias="isVariance")
    variance_kind: Optional[str] = Field(default=None, alias="varianceKind")
    variance_reason: Optional[str] = Field(default=None, alias="varianceReason")
    reference_unit_price: Optional[str] = Field(default=None, alias="referenceUnitPrice")
    reference_amount: Optional[Amount] = Field(default=None, alias="referenceAmount")
    reference_qty_raw: Optional[str] = Field(default=None, alias="referenceQtyRaw")
    reference_qty_normalized: Optional[float] = Field(default=None, alias="referenceQtyNormalized")
    # Return fields
    is_return: Optional[bool] = Field(default=None, alias="isReturn")
    reason_code: Optional[str] = Field(default=None, alias="reasonCode")
    # Rebate fields
    is_rebate: Optional[bool] = Field(default=None, alias="isRebate")
    agreement_id: Optional[str] = Field(default=None, alias="agreementId")
    period_start: Optional[str] = Field(default=None, alias="periodStart")
    period_end: Optional[str] = Field(default=None, alias="periodEnd")
    # Parsing confidence
    confidence_score: Optional[float] = Field(default=None, alias="confidenceScore")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class UpdateLineItemDTO(BaseModel):
    line_number_in_source: Optional[int] = Field(default=None, alias="lineNumberInSource")
    description: Optional[str] = Field(default=None, alias="description")
    qty_raw: Optional[str] = Field(default=None, alias="qtyRaw")
    qty_normalized: Optional[float] = Field(default=None, alias="qtyNormalized")
    qty_unit: Optional[str] = Field(default=None, alias="qtyUnit")
    unit_price: Optional[str] = Field(default=None, alias="unitPrice")
    unit_price_currency: Optional[str] = Field(default=None, alias="unitPriceCurrency")
    amount: Optional[Amount] = Field(default=None, alias="amount")
    sku: Optional[str] = Field(default=None, alias="sku")
    item_code: Optional[str] = Field(default=None, alias="itemCode")
    parsed_raw: Optional[Dict[str, Any]] = Field(default=None, alias="parsedRaw")
    metadata: Optional[Dict[str, Any]] = Field(default=None, alias="metadata")
    # Variance fields
    is_variance: Optional[bool] = Field(default=None, alias="isVariance")
    variance_kind: Optional[str] = Field(default=None, alias="varianceKind")
    variance_reason: Optional[str] = Field(default=None, alias="varianceReason")
    reference_unit_price: Optional[str] = Field(default=None, alias="referenceUnitPrice")
    reference_amount: Optional[Amount] = Field(default=None, alias="referenceAmount")
    reference_qty_raw: Optional[str] = Field(default=None, alias="referenceQtyRaw")
    reference_qty_normalized: Optional[float] = Field(default=None, alias="referenceQtyNormalized")
    # Return fields
    is_return: Optional[bool] = Field(default=None, alias="isReturn")
    reason_code: Optional[str] = Field(default=None, alias="reasonCode")
    # Rebate fields
    is_rebate: Optional[bool] = Field(default=None, alias="isRebate")
    agreement_id: Optional[str] = Field(default=None, alias="agreementId")
    period_start: Optional[str] = Field(default=None, alias="periodStart")
    period_end: Optional[str] = Field(default=None, alias="periodEnd")
    # Parsing confidence
    confidence_score: Optional[float] = Field(default=None, alias="confidenceScore")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True



class LineItemSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "accounting/line-items", header_builder,
                         renew_token, LineItemAPIDTO.parse_obj(data))


class LineItemAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "accounting/line-items", header_builder,
                         renew_token, LineItemAPIDTO.parse_obj(data))


class LineItemsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=LineItemSync,
                         retrieve_data_model=LineItemAPIDTO,
                         create_data_model=CreateLineItemDTO,
                         update_data_model=UpdateLineItemDTO,
                         resource="accounting/line-items")


class LineItemsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=LineItemAsync,
                         retrieve_data_model=LineItemAPIDTO,
                         create_data_model=CreateLineItemDTO,
                         update_data_model=UpdateLineItemDTO,
                         resource="accounting/line-items")




from pydantic import BaseModel, Field
from typing import Optional, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class ActionableAPIDTO(BaseModel):
    id: str = Field(alias="id")
    execution_id: Optional[str] = Field(alias="executionId", default=None)
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    
    assigned_to: Optional[str] = Field(alias="assignedTo", default=None)
    status: str = Field(alias="status")
    priority: int = Field(alias="priority")
    
    title: Optional[str] = Field(alias="title", default=None)
    message: Optional[str] = Field(alias="message", default=None)
    
    context: Dict[str, Any] = Field(alias="context")
    
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateActionableRequest(BaseModel):
    execution_id: Optional[str] = Field(alias="executionId", default=None)
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    
    assigned_to: Optional[str] = Field(alias="assignedTo", default=None)
    status: Optional[str] = Field(alias="status", default="pending")
    priority: Optional[int] = Field(alias="priority", default=0)
    
    title: Optional[str] = Field(alias="title", default=None)
    message: Optional[str] = Field(alias="message", default=None)
    
    context: Optional[Dict[str, Any]] = Field(alias="context", default={})

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateActionableRequest(BaseModel):
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    assigned_to: Optional[str] = Field(alias="assignedTo", default=None)
    status: Optional[str] = Field(alias="status", default=None)
    priority: Optional[int] = Field(alias="priority", default=None)
    title: Optional[str] = Field(alias="title", default=None)
    message: Optional[str] = Field(alias="message", default=None)
    context: Optional[Dict[str, Any]] = Field(alias="context", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ActionableSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "actionables", header_builder, renew_token, ActionableAPIDTO.parse_obj(data))


class ActionableAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "actionables", header_builder, renew_token, ActionableAPIDTO.parse_obj(data))


class ActionablesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, sync_resource=ActionableSync, retrieve_data_model=ActionableAPIDTO,
                         create_data_model=CreateActionableRequest, update_data_model=UpdateActionableRequest, resource="actionables")


class ActionablesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, async_resource=ActionableAsync, retrieve_data_model=ActionableAPIDTO,
                         create_data_model=CreateActionableRequest, update_data_model=UpdateActionableRequest, resource="actionables")



from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class ChannelConnectorAPIDTO(BaseModel):
    id: str = Field(alias="id")
    channel: str = Field(alias="channel")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ChannelConnectorCreate(BaseModel):
    channel: str = Field(alias="channel")
    url: str = Field(alias="url")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ChannelConnectorUpdate(BaseModel):
    url: str = Field(alias="url")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ChannelConnectorSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/channel-connectors", header_builder, renew_token,
                         ChannelConnectorAPIDTO.parse_obj(data))


class ChannelConnectorAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/channel-connectors", header_builder, renew_token,
                         ChannelConnectorAPIDTO.parse_obj(data))


class ChannelConnectorSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, sync_resource=ChannelConnectorSync, retrieve_data_model=ChannelConnectorAPIDTO,
                         create_data_model=ChannelConnectorCreate, update_data_model=ChannelConnectorUpdate,
                         resource="/conversational/channel-connectors")


class ChannelConnectorAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, async_resource=ChannelConnectorAsync, retrieve_data_model=ChannelConnectorAPIDTO,
                         create_data_model=ChannelConnectorCreate, update_data_model=ChannelConnectorUpdate,
                         resource="/conversational/channel-connectors")


from typing import List, Optional

import httpx
from pydantic import BaseModel, Field

from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.helpers import build_headers

class CategoryDTO(BaseModel):
    id: str = Field(alias="id")
    category_key: str = Field(alias="categoryKey")
    label: str = Field(alias="label")
    created_at: str = Field(alias="createdAt")
    updated_at: str = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class CategoryValueDTO(BaseModel):
    id: str = Field(alias="id")
    category_id: str = Field(alias="categoryId")
    category_key: str = Field(alias="categoryKey")
    value: str = Field(alias="value")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class EntityCategoryValueDTO(BaseModel):
    category_id: str = Field(alias="categoryId")
    category_key: str = Field(alias="categoryKey")
    category_value_id: str = Field(alias="categoryValueId")
    entity_type: str = Field(alias="entityType")
    entity_id: str = Field(alias="entityId")
    created_at: str = Field(alias="createdAt")
    updated_at: str = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class NewCategoryDTO(BaseModel):
    category_key: str = Field(alias="categoryKey")
    label: str = Field(alias="label")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class NewCategoryValueDTO(BaseModel):
    id: Optional[str] = Field(alias="id", default=None)
    value: str = Field(alias="value")

class EntityWrapper(BaseModel):
    entity_type: str = Field(alias="entityType")
    entity_id: str = Field(alias="entityId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CategoryBase:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def _get_category_values_url(self, category_id: str) -> str:
        return f"{self.base_url}/v1/category/{category_id}/values"

    def _add_category_value_url(self, category_id: str):
        return f"{self.base_url}/v1/category/{category_id}/value"

    def _delete_category_value_url(self, category_id: str, category_value_id: str):
        return f"{self.base_url}/v1/category/{category_id}/value/{category_value_id}"


class CategoryAsyncModule:
    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401_async
    async def create(self, new_category: dict, values: Optional[List[NewCategoryValueDTO]] = None):
        if values is None:
            values = []
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                "/v1/category",
                headers=self.build_headers(),
                json=NewCategoryDTO.parse_obj(new_category).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            cat_id = response.json()["id"]
            if len(values) > 0:
                category = await self.retrieve(cat_id)
                for value in values:
                    await category.create_category_value(value)

            return cat_id

    @retry_on_401_async
    async def retrieve(self, category_id):
        base_url = self.altscore_client._borrower_central_base_url
        async with httpx.AsyncClient(base_url=base_url) as client:
            response = await client.get(
                "/v1/category/{}".format(category_id),
                headers=self.build_headers(),
                timeout=120
            )
            if response.status_code == 200:
                return CategoryAsync(
                    base_url=base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    category_data=CategoryDTO.parse_obj(response.json())
                )
            if response.status_code == 404:
                return None
            raise_for_status_improved(response)

    @retry_on_401_async
    async def delete(self, category_id):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.delete(
                "/v1/category/{}".format(category_id),
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)

    @retry_on_401_async
    async def set_category_value_to_entity(self, entity: EntityWrapper, category_key: str, category_value_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                "/v1/category/commands/categorize-entity",
                headers=self.build_headers(),
                json={
                    "categoryKey": category_key,
                    "categoryValueId": category_value_id,
                    "entityType": entity.entity_type,
                    "entityId": entity.entity_id,
                }
            )
            raise_for_status_improved(response)
            return None

    @retry_on_401_async
    async def delete_category_value_from_entity(self, entity: EntityWrapper, category_key: str, category_value_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                "/v1/category/commands/delete-entity-category",
                headers=self.build_headers(),
                json={
                    "categoryKey": category_key,
                    "categoryValueId": category_value_id,
                    "entityType": entity.entity_type,
                    "entityId": entity.entity_id,
                }
            )
            raise_for_status_improved(response)
            return None

    @retry_on_401_async
    async def get_category_values_by_entity(self, entity: EntityWrapper):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/category/queries/entity/{entity.entity_type}/{entity.entity_id}",
                headers=self.build_headers(),
            )
            raise_for_status_improved(response)
            return [
                EntityCategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

    @retry_on_401_async
    async def get_category_values_by_entity_type(self, entity_type: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/category/queries/entity/{entity_type}",
                headers=self.build_headers(),
            )
            raise_for_status_improved(response)
            return [
                EntityCategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

    @retry_on_401_async
    async def get_category_values_by_key(self, category_key: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/category/by-key/{category_key}/values",
                headers=self.build_headers(),
            )
            raise_for_status_improved(response)
            return [
                CategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

class CategorySyncModule:
    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401
    def create(self, new_category: dict, values: Optional[List[NewCategoryValueDTO]] = None):
        if values is None:
            values = []
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                "/v1/category",
                headers=self.build_headers(),
                json=NewCategoryDTO.parse_obj(new_category).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)

            cat_id = response.json()["id"]
            if len(values) > 0:
                category = self.retrieve(cat_id)
                for value in values:
                    category.create_category_value(value)

            return cat_id

    @retry_on_401
    def retrieve(self, category_id):
        base_url = self.altscore_client._borrower_central_base_url
        with httpx.Client(base_url=base_url) as client:
            response = client.get(
                "/v1/category/{}".format(category_id),
                headers=self.build_headers(),
                timeout=120
            )
            if response.status_code == 200:
                return CategorySync(
                    base_url=base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    category_data=CategoryDTO.parse_obj(response.json())
                )
            if response.status_code == 404:
                return None
            raise_for_status_improved(response)

    @retry_on_401
    def delete(self, category_id):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.delete(
                "/v1/category/{}".format(category_id),
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)

    @retry_on_401
    def set_category_value_to_entity(self, entity: EntityWrapper, category_key: str, category_value_id: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                "/v1/category/commands/categorize-entity",
                headers=self.build_headers(),
                json={
                    "categoryKey": category_key,
                    "categoryValueId": category_value_id,
                    "entityType": entity.entity_type,
                    "entityId": entity.entity_id,
                }
            )
            raise_for_status_improved(response)
            return None

    @retry_on_401
    def delete_category_value_from_entity(self, entity: EntityWrapper, category_key: str, category_value_id: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                "/v1/category/commands/delete-entity-category",
                headers=self.build_headers(),
                json={
                    "categoryKey": category_key,
                    "categoryValueId": category_value_id,
                    "entityType": entity.entity_type,
                    "entityId": entity.entity_id,
                }
            )
            raise_for_status_improved(response)
            return None

    @retry_on_401
    def get_category_values_by_entity(self, entity: EntityWrapper):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/category/queries/entity/{entity.entity_type}/{entity.entity_id}",
                headers=self.build_headers(),
            )
            raise_for_status_improved(response)
            return [
                EntityCategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

    @retry_on_401
    def get_category_values_by_entity_type(self, entity_type: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/category/queries/entity/{entity_type}",
                headers=self.build_headers(),
            )
            raise_for_status_improved(response)
            return [
                EntityCategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

    @retry_on_401
    def get_category_values_by_key(self, category_key: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/category/by-key/{category_key}/values",
                headers=self.build_headers(),
            )
            raise_for_status_improved(response)
            return [
                CategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

class CategoryAsync(CategoryBase):
    data: CategoryDTO

    def __init__(self, base_url, header_builder, renew_token, category_data: CategoryDTO):
        super().__init__(base_url)
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = category_data

    @retry_on_401_async
    async def create_category_value(self, value: NewCategoryValueDTO):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                self._add_category_value_url(self.data.id),
                headers=self._header_builder(),
                json=value.dict(by_alias=True),
            )
            raise_for_status_improved(response)

    @retry_on_401_async
    async def get_category_values(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.get(
                self._get_category_values_url(self.data.id),
                headers=self._header_builder(),
            )
            raise_for_status_improved(response)
            return [
                CategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

    @retry_on_401_async
    async def delete_category_value(self, category_value_id: str):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.delete(
                self._delete_category_value_url(self.data.id, category_value_id),
                headers=self._header_builder(),
            )
            raise_for_status_improved(response)



class CategorySync(CategoryBase):
    data: CategoryDTO

    def __init__(self, base_url, header_builder, renew_token, category_data: CategoryDTO):
        super().__init__(base_url)
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = category_data

    @retry_on_401
    def create_category_value(self, value: NewCategoryValueDTO):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                self._add_category_value_url(self.data.id),
                headers=self._header_builder(),
                json=value.dict(by_alias=True),
            )
            raise_for_status_improved(response)

    @retry_on_401
    def get_category_values(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.get(
                self._get_category_values_url(self.data.id),
                headers=self._header_builder(),
            )
            raise_for_status_improved(response)
            return [
                CategoryValueDTO.parse_obj(data)
                for data in response.json()
            ]

    @retry_on_401
    def delete_category_value(self, category_value_id: str):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.delete(
                self._delete_category_value_url(self.data.id, category_value_id),
                headers=self._header_builder(),
            )
            raise_for_status_improved(response)



from typing import List, Dict, Optional, Union

import httpx
from pydantic import BaseModel, Field

from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.helpers import build_headers

class DateInterval(BaseModel):
    from_: Optional[str] = Field(alias="from", default=None)
    to: Optional[str] = Field(alias="to", default=None)

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True
        populate_by_name = True

class Filter(BaseModel):
    key: str = Field(alias="key")
    entity_type: str = Field(alias="entityType")
    operator: str = Field(alias="operator")
    value: Union[str | List[str]] = Field(alias="value")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True
        populate_by_name = True


class QueryMetadataDTO(BaseModel):
    intl: Dict[str, Dict] = Field(alias="intl")
    data_type: str = Field(alias="dataType")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class NewQueryDTO(BaseModel):
    metric_key: str = Field(alias="metricKey")
    columns: List[str] = Field(alias="columns")
    metadata_cols: List[str] = Field(alias="metadataCols")
    query_metadata: QueryMetadataDTO = Field(alias="queryMetadata")
    query_template: str = Field(alias="queryTemplate")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class ExecuteQueryDTO(BaseModel):
    metric_key: str = Field(alias="metricKey")
    filters: Optional[List[Filter]] = Field(alias="filters", default=[])
    cohort_date: DateInterval = Field(alias="cohortDate", default=DateInterval())
    analysis_date: Optional[str] = Field(alias="analysisDate", default=None)

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True
        populate_by_name = True

class AnalyticsAsyncModule:
    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401_async
    async def create_query(self, new_metric: Dict):
        base_url = self.altscore_client._borrower_central_base_url
        async with httpx.AsyncClient(base_url=base_url) as client:
            response = await client.post(
                "/v1/analytics/commands/new-metric",
                headers=self.build_headers(),
                json=NewQueryDTO.parse_obj(new_metric).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)

    @retry_on_401_async
    async def execute_query(self, query: Dict):
        base_url = self.altscore_client._borrower_central_base_url
        async with httpx.AsyncClient(base_url=base_url) as client:
            response = await client.post(
                "/v1/analytics/commands/get-metrics",
                headers=self.build_headers(),
                json=ExecuteQueryDTO.parse_obj(query).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            return response.json()

    @retry_on_401_async
    async def delete_query(self, query_key: str):
        base_url = self.altscore_client._borrower_central_base_url
        async with httpx.AsyncClient(base_url=base_url) as client:
            response = await client.delete(
                f"/v1/analytics/query/{query_key}",
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)

class AnalyticsSyncModule:
    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401
    def create_query(self, new_metric: Dict):
        base_url = self.altscore_client._borrower_central_base_url
        with httpx.Client(base_url=base_url) as client:
            response = client.post(
                "/v1/analytics/commands/new-metric",
                headers=self.build_headers(),
                json=NewQueryDTO.parse_obj(new_metric).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)

    @retry_on_401
    def execute_query(self, query: Dict):
        base_url = self.altscore_client._borrower_central_base_url
        with httpx.Client(base_url=base_url) as client:
            response = client.post(
                "/v1/analytics/commands/get-metrics",
                headers=self.build_headers(),
                json=ExecuteQueryDTO.parse_obj(query).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            return response.json()

    @retry_on_401
    def delete_query(self, query_key: str):
        base_url = self.altscore_client._borrower_central_base_url
        with httpx.Client(base_url=base_url) as client:
            response = client.delete(
                f"/v1/analytics/query/{query_key}",
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)